"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.createError = exports.route = exports.router = exports.notFound = exports.sendError = exports.send = exports.json = exports.text = exports.buffer = exports.getHeader = exports.serve = void 0;
const content_type_1 = __importDefault(require("content-type"));
const path_to_regexp_1 = require("path-to-regexp");
const raw_body_1 = __importDefault(require("raw-body"));
const stream_1 = require("stream");
const url_1 = require("url");
const IS_DEV = process.env.NODE_ENV === 'development';
function serve(handler, options = {}) {
    return async function (req, res) {
        try {
            const serverRequest = requestFromHTTP(req, options);
            const serverResponse = responseFromHTTP(res);
            await handler(serverRequest, serverResponse, async (_, res) => notFound(res));
        }
        catch (error) {
            if (options.onError)
                await options.onError(error);
            else if (!res.writableEnded)
                sendError(res, error);
        }
    };
}
exports.serve = serve;
// Request ---------------------------------------------------------------------
function requestFromHTTP(req, options) {
    const originalURL = req.url;
    const serverRequest = Object.defineProperties(req, {
        protocol: cachedGetter(req, () => {
            const socketProtocol = Boolean(req.socket.encrypted) ? 'https' : 'http';
            if (!options.trustProxy)
                return socketProtocol;
            const headerProtocol = getHeader(serverRequest, 'x-forwarded-proto') ?? socketProtocol;
            const commaIndex = headerProtocol.indexOf(',');
            return commaIndex === -1 ? headerProtocol.trim() : headerProtocol.substring(0, commaIndex).trim();
        }),
        query: cachedGetter(req, () => {
            return Object.fromEntries(serverRequest.url.searchParams);
        }),
        url: cachedGetter(req, () => {
            return new url_1.URL(originalURL, `http://${req.headers.host}`);
        }),
    });
    return serverRequest;
}
function getHeader(req, header) {
    const value = req.headers[header];
    return Array.isArray(value) ? value[0] : value;
}
exports.getHeader = getHeader;
const requestBodyMap = new WeakMap();
async function buffer(req, { limit = '1mb', encoding } = {}) {
    const type = req.headers['content-type'] ?? 'text/plain';
    const length = req.headers['content-length'];
    if (encoding === undefined) {
        encoding = content_type_1.default.parse(type).parameters.charset;
    }
    const existingBody = requestBodyMap.get(req);
    if (existingBody)
        return existingBody;
    try {
        const body = Buffer.from(await (0, raw_body_1.default)(req, { limit, length, encoding }));
        requestBodyMap.set(req, body);
        return body;
    }
    catch (error) {
        if (error.type === 'entity.too.large') {
            throw createError(413, `Body exceeded ${limit} limit`, error);
        }
        throw createError(400, 'Invalid body', error);
    }
}
exports.buffer = buffer;
async function text(req, options = {}) {
    return await buffer(req, options).then((body) => body.toString());
}
exports.text = text;
async function json(req, options = {}) {
    return await text(req, options).then((body) => {
        try {
            return JSON.parse(body);
        }
        catch (error) {
            throw createError(400, 'Invalid JSON', error);
        }
    });
}
exports.json = json;
// Response --------------------------------------------------------------------
function responseFromHTTP(res) {
    const serverResponse = Object.defineProperties(res, {});
    return serverResponse;
}
function send(res, code, body = null) {
    res.statusCode = code;
    if (body === null || body === undefined) {
        res.end();
        return;
    }
    if (body instanceof stream_1.Stream || isReadableStream(body)) {
        if (!res.getHeader('Content-Type')) {
            res.setHeader('Content-Type', 'application/octet-stream');
        }
        body.pipe(res);
        return;
    }
    if (Buffer.isBuffer(body)) {
        if (!res.getHeader('Content-Type')) {
            res.setHeader('Content-Type', 'application/octet-stream');
        }
        res.setHeader('Content-Length', body.length);
        res.end(body);
        return;
    }
    let stringifiedBody;
    if (typeof body === 'object' || typeof body === 'number') {
        stringifiedBody = JSON.stringify(body);
        if (!res.getHeader('Content-Type')) {
            res.setHeader('Content-Type', 'application/json; charset=utf-8');
        }
    }
    else {
        stringifiedBody = body;
    }
    res.setHeader('Content-Length', Buffer.byteLength(stringifiedBody));
    res.end(stringifiedBody);
}
exports.send = send;
function sendError(res, error) {
    const statusCode = error.statusCode;
    const message = statusCode ? error.message : 'Internal Server Error';
    send(res, statusCode ?? 500, IS_DEV ? error.stack : message);
    console.error(error.stack);
}
exports.sendError = sendError;
function notFound(res) {
    send(res, 404, 'Not Found');
}
exports.notFound = notFound;
// Router ----------------------------------------------------------------------
const notFoundMiddleware = async (_, res) => send(res, 404, 'Not Found');
function router(...middleware) {
    return async function (req, res) {
        const next = async (req, res, idx) => {
            const current = middleware[idx] ?? notFoundMiddleware;
            await current(req, res, (req, res) => next(req, res, idx + 1));
        };
        await next(req, res, 0);
    };
}
exports.router = router;
// Implementation
function route(method, path, handler, validator) {
    const matchPath = (0, path_to_regexp_1.match)(path);
    const compilePath = (0, path_to_regexp_1.compile)(path);
    const routeHandler = async (req, res, next) => {
        if (req.method !== method)
            return await next(req, res);
        const pathMatch = matchPath(req.url.pathname);
        if (!pathMatch)
            return await next(req, res);
        req.params = pathMatch.params;
        let body = undefined;
        if (validator) {
            body = await json(req);
            if (typeof body !== 'object' || body === null || !validator(body)) {
                return sendError(res, createError(422, 'Request body failed validation'));
            }
        }
        const responseBody = await Promise.resolve(handler(Object.assign(req, { body }), res, next));
        if (responseBody === null) {
            send(res, 204, null);
            return;
        }
        if (responseBody !== undefined) {
            send(res, res.statusCode ?? 200, responseBody);
        }
    };
    return Object.assign(routeHandler, { method, route: path, compilePath });
}
exports.route = route;
function createError(code, message, original) {
    const error = new Error(message);
    error.statusCode = code;
    error.originalError = original;
    return error;
}
exports.createError = createError;
// Utilities -------------------------------------------------------------------
function isStream(val) {
    return val !== null && typeof val === 'object' && typeof val.pipe === 'object';
}
function isReadableStream(val) {
    return (isStream(val) &&
        val.readable !== false &&
        typeof val._read === 'function' &&
        typeof val._readableState === 'object');
}
function cachedGetter(obj, getter) {
    const cache = new WeakMap();
    return {
        get: () => {
            if (cache.has(obj))
                return cache.get(obj);
            const value = getter();
            cache.set(obj, value);
            return value;
        },
        enumerable: true,
    };
}
//# sourceMappingURL=zap.js.map